from bpy.types import WorkSpaceTool, Operator
from pathlib import Path
import bpy
from . import vox_manager
from . import vox_preview_voxel_manager
from mathutils import Vector
from bl_ui.space_toolsystem_common import ToolSelectPanelHelper


class color_voxel_tool(WorkSpaceTool):
    """Change color of voxels under cursor"""
    bl_space_type = 'VIEW_3D'
    bl_context_mode = 'EDIT_MESH'
    bl_idname = "mesh.vox_color_voxel_tool"
    bl_label = "Paint voxel"
    bl_description = (
        "Vox\n"
        "Change color of voxels under cursor"
    )
    bl_cursor = "PAINT_BRUSH"
    bl_icon = (Path(__file__).parent / "" / "ops.mesh.color_voxel").as_posix()
    bl_widget = "EDIT_MESH_GGT_vox_tool"
    bl_keymap = (
        ("mesh.vox_color_voxel", {"type": 'LEFTMOUSE', "value": 'ANY'},
         {"properties": []}),
        ("mesh.vox_prview_voxel", {"type": 'MOUSEMOVE', "value": 'ANY'},
         {"properties": [("add_new", False)]}),
        ("mesh.vox_get_voxel_color",
         {"type": 'LEFTMOUSE', "alt": True, "value": 'ANY'},
         {"properties": []}),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("mesh.vox_color_voxel")
        vs = context.object.vox_settings
        active = vs.active_slot

        layout.prop(props, "operatormode")

        row = layout.row(align=True)

        row.prop(props, "apply_color", icon="BRUSH_DATA", icon_only=True)
        row.prop(props, "apply_color", icon="META_CUBE", icon_only=True, invert_checkbox=True)

        if props.apply_color:
            layout.prop(props, "color", icon_only=True)
        else:
            if vs.vox_slots[active].vox_collection is None:
                text = "Basic blocks"
            else:
                text = vs.vox_slots[active].vox_collection.name
            layout.popover(
                "VOX_PT_slot_selector",
                text=text,
                text_ctxt='',
                icon='OUTLINER_COLLECTION',
                icon_value=0
            )
        if props.operatormode == "BRUSH":
            layout.prop(props, "radius")
        if props.apply_color:
            layout.prop(props, "strength")

        if context.region.type == "TOOL_HEADER":
            layout.popover("VIEW3D_PT_vox_color_voxel_settings",
                           text="Settings")
        else:
            if props.operatormode == "BRUSH":
                layout.prop(props, "falloff")
            layout.prop(props, "spacing")
            layout.prop(props, "accumulate")


class VIEW3D_PT_vox_color_voxel_settings(bpy.types.Panel):
    bl_label = "1Advanced"
    bl_options = {'DEFAULT_CLOSED'}
    bl_ui_units_x = 14
    bl_space_type = 'VIEW_3D'
    bl_region_type = "TOOL_HEADER"

    def draw(self, context):
        layout = self.layout

        layout.use_property_split = True
        layout.use_property_decorate = False  # No animation.
        tool = ToolSelectPanelHelper.tool_active_from_context(context)
        props = tool.operator_properties("mesh.vox_color_voxel")

        if props.operatormode == "BRUSH":
            layout.prop(props, "falloff")
        layout.prop(props, "spacing")
        layout.prop(props, "accumulate")


class color_voxel_op(Operator):
    """"Change color of voxels under cursor"""
    bl_idname = "mesh.vox_color_voxel"
    bl_label = "Paint voxels"
    bl_options = {'REGISTER', 'UNDO'}

    apply_color:  bpy.props.BoolProperty(name="Apply Color", default=True)

    apply_random:  bpy.props.BoolProperty(name="Rebuild part of the model", default=False)

    color: bpy.props.FloatVectorProperty(
        name="Color", subtype="COLOR", min=0, max=1, default=[1.0, 1.0, 1.0])

    strength: bpy.props.FloatProperty(
        name="Strength", min=0, max=1, default=1.0)

    spacing: bpy.props.FloatProperty(
        name="Spacing", min=0, max=1, default=0.1)

    falloff: bpy.props.FloatProperty(
        name="Falloff", min=0, max=1, default=1.0)

    operatormode: bpy.props.EnumProperty(
        name='Mode',
        description='Pain tool',
        items={
            ('DEF', 'Per Voxel', 'Draw only voxel under cursor'),
            ('BRUSH', 'Brush', 'Paint with 3D-sphere with given radius and center at point under cusor')},
        default='DEF')

    radius: bpy.props.IntProperty(
        name="Radius", min=0, max=500, default=10)

    oldPosition = Vector((0, 0))

    accumulate: bpy.props.BoolProperty(name="Accumulate", default=False)

    @ classmethod
    def poll(cls, context):
        return (context.mode == 'EDIT_MESH')

    def perfomDraw(self, context, event):

        location, vect = vox_manager.get_addvox_position(
            context, event.mouse_region_x, event.mouse_region_y, False,
            self.operatormode == 'BRUSH')
        if not location:
            return False

        if(self.operatormode == 'BRUSH'):
            location2 = vox_manager.get_plane_intersection(
                context, event.mouse_region_x + self.radius, event.mouse_region_y, location, True)
            location3 = vox_manager.get_plane_intersection(
                context, event.mouse_region_x, event.mouse_region_y + self.radius, location, True)
            distance1 = location - location2
            distance2 = location - location3
            distance = min(distance1.length, distance2.length)

            if self.apply_color:
                vox_manager.set_color_inradius(
                    context, location, self.color, self.strength, distance, self.falloff)
            else:
                vox_manager.set_model_inradius(context, location, distance)

        else:
            if self.apply_color:
                vox_manager.set_color(
                    context, location, self.color, self.strength)
            else:
                vox_manager.set_model(context, location)
        return True

    def invoke(self, context, event):
        if event.value != "PRESS":
            return {'CANCELLED'}
        self.oldPosition = Vector((event.mouse_region_x, event.mouse_region_y))
        vox_manager.set_color_prep(context)

        if event.value == "PRESS":
            self.perfomDraw(context, event)

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        return {'PASS_THROUGH'}

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':
            radius = self.radius if self.operatormode == 'BRUSH' else 25
            spacing = self.spacing * radius
            diff = Vector(
                (event.mouse_region_x, event.mouse_region_y)) - self.oldPosition
            if diff.length > spacing:
                self.oldPosition = Vector(
                    (event.mouse_region_x, event.mouse_region_y))
                self.perfomDraw(context, event)
            return {'PASS_THROUGH'}

        elif event.type == 'LEFTMOUSE':
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}
