from bpy.types import WorkSpaceTool, Operator
from pathlib import Path
import bpy
from mathutils import Vector
from . import vox_manager
from . import vox_preview_voxel_manager
from . import vox_gn_setup


class draw_voxel_tool(WorkSpaceTool):
    """Adds new voxels under cursor"""
    bl_space_type = 'VIEW_3D'
    bl_context_mode = 'EDIT_MESH'
    bl_idname = "mesh.vox_draw_voxel_tool"
    bl_label = "Draw voxel"
    bl_description = (
        "Vox\n"
        "Add new voxels"
    )
    bl_icon = (Path(__file__).parent / "" / "ops.mesh.draw_voxel").as_posix()
    bl_widget = "EDIT_MESH_GGT_vox_tool"
    bl_keymap = (
        ("mesh.vox_draw_voxel",
         {"type": 'LEFTMOUSE', "value": 'ANY'},
         {"properties": []}),
        ("mesh.vox_draw_voxel",
         {"type": 'LEFTMOUSE', "ctrl": True, "value": 'ANY'},
         {"properties": [("remove", True)]}),
        ("mesh.vox_prview_voxel",
         {"type": 'MOUSEMOVE', "value": 'ANY'},
         {"properties": [("add_new", True)]}),
        ("mesh.vox_prview_voxel",
         {"type": 'MOUSEMOVE', "ctrl": True, "value": 'ANY'},
         {"properties": [("add_new", False), ("tool_color", (1, 0, 0))]}),
        ("mesh.vox_get_voxel_color",
         {"type": 'LEFTMOUSE', "alt": True, "value": 'ANY'},
         {"properties": []}),
    )

    def draw_settings(context, layout, tool):
        props = tool.operator_properties("mesh.vox_draw_voxel")
        vs = context.object.vox_settings
        active = vs.active_slot

        if len(vs.vox_slots) == 0:
            layout.operator("object.vox_rebuild_gn_tree")
        else:
            layout.prop(props, "operatormode")
            if vs.vox_slots[active].vox_collection is None:
                text = "Basic blocks"
            else:
                text = vs.vox_slots[active].vox_collection.name
            layout.popover(
                "VOX_PT_slot_selector",
                text=text,
                text_ctxt='',
                icon='OUTLINER_COLLECTION',
                icon_value=0
            )

            layout.prop(props, "color", icon_only=True)


class draw_voxel_op(Operator):
    """Add new voxels"""
    bl_idname = "mesh.vox_draw_voxel"
    bl_label = "Draw voxels"
    bl_options = {'REGISTER', 'UNDO'}

    color: bpy.props.FloatVectorProperty(
        name="Color", subtype="COLOR", min=0, max=1, default=[1.0, 1.0, 1.0])

    remove: bpy.props.BoolProperty(name='Remove', default=False)
    pointlocation: Vector((0, 0, 0))
    startlocation: Vector((0, 0, 0))
    secondlocation: Vector((0, 0, 0))
    step: 0
    start_mousepos: Vector((0, 0))
    zextrude: 0
    drawnormal: Vector((0, 0, 0))

    operatormode: bpy.props.EnumProperty(
        name='Mode',
        description='Style of adding voxels',
        items={
            ('DEF', 'Default', 'Click to add, click and drag to add a rect'),
            ('EXTRUDE', 'Draw and extrude', 'Draw a rect and extrude it')},
        default='DEF')

    @ classmethod
    def poll(cls, context):
        return (context.mode == 'EDIT_MESH')

    def invoke(self, context, event):
        if event.value != "PRESS":
            return {'CANCELLED'}
        vs = context.object.vox_settings
        if len(vs.vox_slots) == 0:
            bpy.ops.object.vox_rebuild_gn_tree()

        location, vect = vox_manager.get_addvox_position(
            context, event.mouse_region_x, event.mouse_region_y, not self.remove)
        exlocation, vect = vox_manager.get_addvox_position(
            context, event.mouse_region_x, event.mouse_region_y, not self.remove, True)
        if not exlocation:
            return {'CANCELLED'}

        self.drawnormal = vect if vect is not False else Vector((0, 0, 1))
        self.startlocation = location
        self.pointlocation = exlocation-location
        self.step = 0
        context.window_manager.modal_handler_add(self)

        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        context.area.tag_redraw()

        if event.type == 'MOUSEMOVE':

            if(self.operatormode == 'EXTRUDE' and self.step == 1):
                self.zextrude = vox_manager.get_height(
                    context, self.start_mousepos,
                    Vector((event.mouse_region_x, event.mouse_region_y)), self.drawnormal)

                location = self.secondlocation + (self.drawnormal * self.zextrude)

            else:
                location = vox_manager.get_plane_intersection(
                    context, event.mouse_region_x, event.mouse_region_y,
                    self.startlocation, False, self.pointlocation, self.drawnormal)

            vox_preview_voxel_manager.setHelper(
                self.startlocation, context, "VOLUME", location,
                (1, 1, 1) if not self.remove else (1, 0, 0))

            return {'RUNNING_MODAL'}

        elif event.type == 'LEFTMOUSE':

            if self.operatormode == 'EXTRUDE' and self.step == 1:
                if event.value == "RELEASE":
                    self.step == 0

                    if not self.remove:
                        vox_manager.add_vertex_volume(
                            context, self.startlocation,
                            self.secondlocation + (self.drawnormal * self.zextrude), self.color)
                    else:
                        vox_manager.remove_vertex_volume(
                            context, self.startlocation,
                            self.secondlocation + (self.drawnormal * self.zextrude))
                    return {'FINISHED'}

                else:
                    return {'RUNNING_MODAL'}

            location = vox_manager.get_plane_intersection(
                context, event.mouse_region_x, event.mouse_region_y,
                self.startlocation, False, self.pointlocation, self.drawnormal)

            if not location:
                return {'FINISHED'}

            if not self.remove:
                if(self.startlocation == location):
                    vox_manager.add_vertex(context, location, self.color)
                else:
                    vox_manager.add_vertex_volume(
                        context, self.startlocation, location, self.color)

            else:
                vox_manager.remove_vertex_volume(
                    context, self.startlocation, location)

            if self.operatormode == 'EXTRUDE':
                self.secondlocation = location
                self.step = 1
                self.start_mousepos = Vector(
                    (event.mouse_region_x, event.mouse_region_y))
                return {'RUNNING_MODAL'}

            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            vox_preview_voxel_manager.resetHelper()
            return {'CANCELLED'}

        return {'RUNNING_MODAL'}
