import bpy
from . import vox_gn_setup
from . import vox_manager
from . import vox_preview_voxel_manager


class VOX_OP_rebuild_model(bpy.types.Operator):
    """Rebuild model in case that blocks has been changed"""
    bl_idname = "object.vox_rebuild_model"
    bl_label = "Rebuild model"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.object is not None

    def execute(self, context):
        vox_manager.rebuild_model(context)
        return {"FINISHED"}


class VOX_PT_slot_selector(bpy.types.Panel):
    """Select slot with models"""
    bl_label = "Active slot"
    bl_idname = "VOX_PT_slot_selector"
    bl_context = "layout"
    bl_options = {'DEFAULT_CLOSED'}
    bl_ui_units_x = 14
    bl_space_type = 'VIEW_3D'
    bl_region_type = "TOOL_HEADER"

    def draw(self, context):
        layout = self.layout
        obj = context.object
        row = layout.row()

        row.template_list("VOX_UL_slots", "", obj.vox_settings, "vox_slots",
                          obj.vox_settings, "active_slot", rows=3)

        col = row.column(align=True)
        col.operator("object.vox_slot_add", icon="ADD", text="")
        col.operator("object.vox_slot_remove", icon="REMOVE", text="")
        col.separator()

        if obj.vox_settings.active_slot < len(obj.vox_settings.vox_slots):
            row = layout.row()
            row.prop(obj.vox_settings.vox_slots[obj.vox_settings.active_slot], "vox_collection")


def update_vox_slot(self, context):
    vox_gn_setup.rebuild_GN_tree(context)
    vox_manager.rebuild_model(context)


def remove_vox_slot(context):
    vs = context.object.vox_settings
    obj = context.object
    if len(vs.vox_slots) > 1:

        origmode = obj.mode
        bpy.ops.object.mode_set(mode='EDIT')

        vox_manager.change_slot(context, vs.active_slot, 0)
        for i in range(vs.active_slot+1, len(vs.vox_slots)):
            vox_manager.change_slot(context, i, i-1)
        vs.vox_slots.remove(vs.active_slot)
        if vs.active_slot >= len(vs.vox_slots):
            vs.active_slot = len(vs.vox_slots)-1

        bpy.ops.object.mode_set(mode=origmode)

    else:
        return False

    vox_gn_setup.rebuild_GN_tree(context)
    return True


class VOX_OP_remove_vox_slot(bpy.types.Operator):
    """Remove Vox slot"""
    bl_idname = "object.vox_slot_remove"
    bl_label = "Remove Vox slot"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.object is not None

    def execute(self, context):
        if not remove_vox_slot(context):
            self.report({'ERROR'}, "Cannot remove last slot")
        return {"FINISHED"}


def add_vox_slot(context):
    vs = context.object.vox_settings
    newslot = vs.vox_slots.add()
    newslot.name = 'Slot'+str(len(vs.vox_slots))
    vox_gn_setup.rebuild_GN_tree(context)
    vs.active_slot = len(vs.vox_slots)-1
    return newslot


class VOX_OP_add_vox_slot(bpy.types.Operator):
    """Add Vox slot"""
    bl_idname = "object.vox_slot_add"
    bl_label = "Add Vox slot"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.object is not None

    def execute(self, context):
        add_vox_slot(context)
        return {"FINISHED"}


def draw_object_settings(self, context):
    layout = self.layout

    obj = context.object

    if obj is None:
        return

    row = layout.row(align=True)
    row.prop(obj.vox_settings, "active", text="Use as voxel", icon="MESH_CUBE")
    row.prop(obj.vox_settings, "active", text="Build a model from voxels", icon="VIEW_ORTHO", invert_checkbox=True)

    if obj.vox_settings.active:
        if context.mode == 'OBJECT':
            vox_preview_voxel_manager.occupied_preview(context)

        for z in range(-1, 2)[::-1]:
            if z == -1:
                layout.label(text="Layer Below", translate=False, icon="NONE")
            elif z == 0:
                layout.label(text="Same Layer", translate=False, icon="NONE")
            else:
                layout.label(text="Layer Above", translate=False, icon="NONE")

            block = layout.column(align=True)
            for y in range(-1, 2)[::-1]:
                row = block.row(align=True)
                for x in range(-1, 2):
                    if x != 0 or y != 0 or z != 0:
                        id = vox_manager.vox_offset_to_id(x, y, z)
                        row.prop(obj.vox_settings, "occupied"+str(id), index=id, text="")  # icon_only=True
                    else:
                        sub = row.row()
                        sub.scale_x = 0.666666
                        sub.label(text="", translate=False, icon="NONE")
                        sub.label(text="", translate=False, icon="MESH_CUBE")
                        sub.label(text="", translate=False, icon="NONE")

        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()

        row = col.row(heading="Use Custom Weight")
        row.prop(obj.vox_settings, "use_custom_weight", text="")
        sub = row.row()
        if obj.vox_settings.use_custom_weight:
            sub.prop(obj.vox_settings, "custom_weight_value", text="")
        else:
            sub.label(text=str(vox_manager.get_model_weigit(obj)), translate=False, icon="NONE")

        col.prop(obj.vox_settings, "random_count")
        col.prop(obj.vox_settings, "allow_rotation")
        col.prop(obj.vox_settings, "only_on_ground")

        row = col.row(heading="Only on even")
        row.prop(obj.vox_settings, "even_x", text="X")
        row.prop(obj.vox_settings, "even_y", text="Y")
        row.prop(obj.vox_settings, "even_z", text="Z")

    else:
        row = layout.row()
        if context.mode == 'OBJECT':
            vox_preview_voxel_manager.resetHelper()
        row.template_list("VOX_UL_slots", "", obj.vox_settings, "vox_slots",
                          obj.vox_settings, "active_slot", rows=3)

        col = row.column(align=True)
        col.operator("object.vox_slot_add", icon="ADD", text="")
        col.operator("object.vox_slot_remove", icon="REMOVE", text="")
        col.separator()

        if obj.vox_settings.active_slot < len(obj.vox_settings.vox_slots):
            row = layout.row()
            row.prop(obj.vox_settings.vox_slots[obj.vox_settings.active_slot], "vox_collection")
        row = layout.row()
        row.operator("object.vox_rebuild_gn_tree")
        row.operator("object.vox_rebuild_model")

        layout.use_property_split = True
        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)
        col = flow.column()
        col.prop(obj.vox_settings, "zero_level_ground")


class VOX_PT_object_settings(bpy.types.Panel):
    """Vox Object Settings"""

    bl_label = "Vox"
    bl_idname = "VOX_PT_object_settings"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    def draw(self, context):
        draw_object_settings(self, context)


class VOX_PT_object_settings_panel(bpy.types.Panel):
    """Vox Object Settings panel"""
    bl_label = "Vox"
    bl_idname = "VOX_PT_object_settings_panel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"

    def draw(self, context):
        draw_object_settings(self, context)


class VOX_UL_slots(bpy.types.UIList):

    def draw_item(self, context, layout, data, item, icon, active_data, active_propname):
        ob = data
        slot = item
        co = slot.vox_collection

        if self.layout_type in {"DEFAULT", "COMPACT"}:

            if co is not None:
                layout.label(text=co.name, translate=False, icon="OUTLINER_COLLECTION")
            else:
                layout.label(text="", translate=False, icon="OUTLINER_COLLECTION")

        elif self.layout_type in {"GRID"}:
            layout.alignment = "CENTER"
            layout.label(text="123", icon="OUTLINER_COLLECTION")


def collection_filter(slot, col):
    obj = bpy.context.object
    return obj.name not in col.all_objects


class vox_slots(bpy.types.PropertyGroup):
    vox_collection: bpy.props.PointerProperty(
        type=bpy.types.Collection, name="Collection", update=update_vox_slot, poll=collection_filter)


class vox_object_settings(bpy.types.PropertyGroup):

    active: bpy.props.BoolProperty(name="Use as block", options=set(), default=True)

    # active block props:
    # siblings: bpy.props.IntVectorProperty(name="Siblings", options=set(),  size=27, min=0, max=3)
    items = [
        ("DEF", '-', "Not Defined - This position doesn't matter", "NONE", 0),
        ("EMPTY", 'Empty or Other', 'This position should not be occupied by same slot',  "REMOVE", 1),
        ("SAME", 'Same Slot', 'This position should be occupied by same slot',  "RADIOBUT_ON", 2),
        ("ANY",  'Any Slot', 'This position should be occupied by any slot', "RADIOBUT_OFF", 3),
        ("FREE", 'Strict Empty', 'This position should be empty',  "X", 4),
    ]

    occupied0: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied1: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied2: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied3: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied4: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied5: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied6: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied7: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied8: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied9: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied10: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied11: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied12: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    # occupied13: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied14: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied15: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied16: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied17: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied18: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied19: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied20: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied21: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied22: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied23: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied24: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied25: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    occupied26: bpy.props.EnumProperty(items=items, name="Siblings", options=set())
    allow_rotation: bpy.props.BoolProperty(
        name="Allow Rotation", description="Allow block to be rotated 90 deg. in Z axis", options=set(), default=True)

    random_count: bpy.props.IntProperty(name="Random probability, if the same weight", min=1, options=set(), default=1)
    only_on_ground: bpy.props.BoolProperty(name="Only on the ground level",  options=set(), default=False)
    even_x: bpy.props.BoolProperty(name="Only on blocks on even x coordinate",  options=set(), default=False)
    even_y: bpy.props.BoolProperty(name="Only on blocks on even y coordinate",  options=set(), default=False)
    even_z: bpy.props.BoolProperty(name="Only on blocks on even z coordinate",  options=set(), default=False)
    use_custom_weight: bpy.props.BoolProperty(name="Use custom weight",  options=set(), default=False)
    custom_weight_value: bpy.props.IntProperty(name="Custom weight", min=0, options=set(), default=0)

    # active disabled props:
    zero_level_ground: bpy.props.BoolProperty(name="Zero level ground", options=set(), default=True)
    active_slot: bpy.props.IntProperty(name="Active vox slot",  options=set(), default=0)
    vox_slots: bpy.props.CollectionProperty(type=vox_slots, options=set(),  name="Vox slot")
